
#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import unittest
import validators
import time

from datetime import datetime

from src.base.utils import SimpleUtil as SU
from src.base.logger import AuraLogger
from src.base.config import AuraConfig
from src.control import EngineExecutor



class TestLogger(unittest.TestCase):
    """
    Testing the Logger.
    """
    aura_logger = None

    def setUp(self):
        self.config = AuraConfig()
        self.aura_logger = AuraLogger(self.config)

    def test_logger(self):
        self.assertTrue(self.aura_logger.logger.hasHandlers())


class TestConfig(unittest.TestCase):
    """
    Testing the Configuration.
    """
    config = None

    def setUp(self):
        self.config = AuraConfig()


    def test_config(self):
        # Check if config is available
        self.assertIsNotNone(self.config.ini_path)

        # Check if "install_dir" is a valid directory (is evaluated at runtime)
        self.assertTrue(os.path.isdir(self.config.get("install_dir")))

        # Check API Urls
        self.assertTrue(validators.url(self.config.get("api_steering_status")))
        self.assertTrue(validators.url(self.config.get("api_steering_calendar")))
        self.assertTrue(validators.url(self.config.get("api_tank_status")))
        tank_playlist_url = self.config.get("api_tank_playlist").replace("${ID}", "1")
        self.assertTrue(validators.url(tank_playlist_url))
        self.assertTrue(validators.url(self.config.get("api_engine_status")))
        self.assertTrue(validators.url(self.config.get("api_engine_store_playlog")))
        self.assertTrue(validators.url(self.config.get("api_engine_store_clock")))
        engine_health_url = self.config.get("api_engine_store_health").replace("${ENGINE_NUMBER}", "1")
        self.assertTrue(validators.url(engine_health_url))

        # Check if Liquidsoap "socket_dir" is set and a directory
        self.assertTrue(os.path.isdir(self.config.get("socket_dir")))

        # Check if database settings are set
        self.assertIsNotNone(self.config.get("db_user"))
        self.assertIsNotNone(self.config.get("db_pass"))
        self.assertIsNotNone(self.config.get("db_name"))
        self.assertIsNotNone(self.config.get("db_host"))



class TestEngineExecutor(unittest.TestCase):
    """
    Testing the EngineExecutor.
    """

    def setUp(self):
        None


    def test_single_executor(self):
        # Initialize stae and executor params
        global_state = ["none"]
        due_time = SU.timestamp() + 2
        def f(param):
            global_state[0] = param

        # Before the executor is done ther should be the initial value
        e = EngineExecutor("RANDOM_NAMESPACE", None, due_time, f, "hello world")
        self.assertEqual("none", global_state[0])
        self.assertNotEqual("hello world", global_state[0])

        # After 3 seconds there should be the updated value
        time.sleep(5)
        self.assertEqual("hello world", global_state[0])




if __name__ == '__main__':
    unittest.main()